@@ -1,6 +1,8 @@
 from __future__ import annotations
 
 import enum
+import sys
+import warnings
 from abc import ABC
 from abc import abstractmethod
 from copy import deepcopy
@@ -9,22 +11,25 @@ from functools import reduce
 from itertools import chain
 from time import monotonic_ns
 from typing import Any
-from typing import Callable
+from typing import cast
+from typing import Dict
 from typing import Literal
 from typing import Protocol
-from typing import TYPE_CHECKING
+from typing import Tuple
 
 import numpy as np
 
-if TYPE_CHECKING:
-    # requires numpy >= 1.21
-    import numpy.typing as npt
+from ._boundaries import BoundaryRegistry
+from ._typing import HCIArray
+from ._typing import RealArray
 
-    from typing import TypeVar
+if sys.version_info >= (3, 9):
+    from collections.abc import Callable
+else:
+    from typing import Callable
 
-    Real = TypeVar("Real", np.float32, np.float64)
-    RealArray = npt.NDArray[Real]
-    HCIArray = npt.NDArray[np.uint16]
+
+BoundarySpec = Tuple[Tuple[str, str, str], ...]
 
 
 class Geometry(enum.Enum):
@@ -36,18 +41,19 @@ class Geometry(enum.Enum):
 
 
 class DepositionMethod(enum.Enum):
-    PARTICLE_IN_CELL = enum.auto()
+    NEAREST_GRID_POINT = enum.auto()
     CLOUD_IN_CELL = enum.auto()
     TRIANGULAR_SHAPED_CLOUD = enum.auto()
 
 
 _deposition_method_names: dict[str, DepositionMethod] = {
-    m.name.lower(): m for m in DepositionMethod
-} | {"".join([w[0] for w in m.name.split("_")]).lower(): m for m in DepositionMethod}
+    **{m.name.lower(): m for m in DepositionMethod},
+    **{"".join([w[0] for w in m.name.split("_")]).lower(): m for m in DepositionMethod},
+}
 
 
 Name = str
-FieldMap = dict[Name, np.ndarray]
+FieldMap = Dict[Name, np.ndarray]
 DepositionMethodT = Callable[
     [
         "RealArray",
@@ -57,6 +63,7 @@ DepositionMethodT = Callable[
         "RealArray",
         "RealArray",
         "RealArray",
+        "RealArray",
         "HCIArray",
         "RealArray",
     ],
@@ -231,6 +238,21 @@ class Grid(CoordinateValidatorMixin):
     def ndim(self) -> int:
         return len(self.axes)
 
+    @property
+    def cell_volumes(self) -> RealArray:
+        """
+        3D: (nx, ny, nz) array representing volumes
+        2D: (nx, ny) array representing surface
+        1D: (nx,) array representing widths (redundant with cell_widths)
+        """
+        widths = list(self.cell_widths.values())
+        if self.geometry is Geometry.CARTESIAN:
+            return cast(RealArray, np.prod(np.meshgrid(*widths), axis=0))
+        else:
+            raise NotImplementedError(
+                f"cell_volumes property is not implemented for {self.geometry}"
+            )
+
 
 class ParticleSet(CoordinateValidatorMixin):
     def __init__(
@@ -274,6 +296,7 @@ class Dataset(ValidatorMixin):
         self.geometry = geometry
         self.grid = grid
         self.particles = particles
+        self.boundary_recipes = BoundaryRegistry()
 
         if self.grid is not None:
             self.axes = self.grid.axes
@@ -285,7 +308,6 @@ class Dataset(ValidatorMixin):
                 "Grid and/or particle data must be provided"
             )
         self.metadata = deepcopy(metadata) if metadata is not None else {}
-        self._cache: dict[tuple[Name, DepositionMethod], np.ndarray] = {}
 
         super().__init__()
 
@@ -384,15 +406,18 @@ class Dataset(ValidatorMixin):
         /,
         *,
         method: Literal[
-            "pic",
-            "particle_in_cell",
+            "ngp",
+            "nearest_grid_point",
             "cic",
             "cloud_in_cell",
             "tsc",
             "triangular_shaped_cloud",
         ],
+        boundaries: dict[Name, tuple[Name, Name]] | None = None,
         verbose: bool = False,
         return_ghost_padded_array: bool = False,
+        weight_field: Name | None = None,
+        weight_field_boundaries: dict[Name, tuple[Name, Name]] | None = None,
     ) -> np.ndarray:
         r"""
         Perform particle deposition and return the result as a grid field.
@@ -402,22 +427,45 @@ class Dataset(ValidatorMixin):
         particle_field_key (positional only): str
            label of the particle field to deposit
 
-        method (keyword only):  'pic', 'cic' or 'tsc'
-           full names ('particle_in_cell', 'cloud_in_cell', and 'triangular_shaped_cloud')
-           are also valid
+        method (keyword only): 'ngp', 'cic' or 'tsc'
+           full names ('nearest_grid_point', 'cloud_in_cell', and
+           'triangular_shaped_cloud') are also valid
 
         verbose (keyword only): bool (default False)
            if True, print execution time for hot loops (indexing and deposition)
 
         return_ghost_padded_array (keyword only): bool (default False)
-           if True, return the complete deposition array, including one extra cell layer
-           per direction and per side. This option is meant as a debugging tool for
-           methods that leak some particle data outside the active domain (cic and tsc).
+           if True, return the complete deposition array, including one extra
+           cell layer per direction and per side. This option is meant as a
+           debugging tool for methods that leak some particle data outside the
+           active domain (cic and tsc).
 
+        weight_field (keyword only): str
+           label of another field to use as weights. Let u be the field to
+           deposit and w be the weight field. Let u' and w' be their equivalent
+           on-grid descriptions. u' is obtained as
+
+           w'(x) = Σ w(i) c(i,x)
+           u'(x) = (1/w'(x)) Σ u(i) w(i) c(i,x)
+
+           where x is the spatial position, i is a particle index, and w(i,x)
+           are geometric coefficients associated with the deposition method.
+
+        boundaries and weigth_field_boundaries (keyword only): dict
+           Maps from axis names (str) to boundary recipe keys (str, str)
+           representing left/right boundaries. By default all axes will use
+           'open' boundaries on both sides. Specifying boundaries for all axes
+           is not mandated, but note that recipes are applied in the order of
+           specified axes (any unspecified axes will be treated last).
+
+           weight_field_boundaries is required if weight field is used in
+           combinations with boundaries.
+
+           Boundary recipes are applied the weight field (if any) first.
         """
-        from .clib._deposition_methods import _deposit_pic_1D  # type: ignore [import]
-        from .clib._deposition_methods import _deposit_pic_2D  # type: ignore [import]
-        from .clib._deposition_methods import _deposit_pic_3D  # type: ignore [import]
+        from .clib._deposition_methods import _deposit_ngp_1D  # type: ignore [import]
+        from .clib._deposition_methods import _deposit_ngp_2D  # type: ignore [import]
+        from .clib._deposition_methods import _deposit_ngp_3D  # type: ignore [import]
         from .clib._deposition_methods import _deposit_cic_1D  # type: ignore [import]
         from .clib._deposition_methods import _deposit_cic_2D  # type: ignore [import]
         from .clib._deposition_methods import _deposit_cic_3D  # type: ignore [import]
@@ -425,6 +473,15 @@ class Dataset(ValidatorMixin):
         from .clib._deposition_methods import _deposit_tsc_2D  # type: ignore [import]
         from .clib._deposition_methods import _deposit_tsc_3D  # type: ignore [import]
 
+        if method in ("pic", "particle_in_cell"):
+            warnings.warn(
+                f"{method=!r} is a deprecated alias for method='ngp', "
+                "please use 'ngp' (or 'nearest_grid_point') directly",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+            method = "ngp"
+
         if method in _deposition_method_names:
             mkey = _deposition_method_names[method]
         else:
@@ -442,24 +499,30 @@ class Dataset(ValidatorMixin):
         if particle_field_key not in self.particles.fields:
             raise ValueError(f"Unknown particle field {particle_field_key!r}")
 
-        if (particle_field_key, mkey) in self._cache:
-            padded_ret_array = self._cache[particle_field_key, mkey]
-            if return_ghost_padded_array:
-                return padded_ret_array
-            elif self.grid.ndim == 1:
-                return padded_ret_array[1:-1]
-            elif self.grid.ndim == 2:
-                return padded_ret_array[1:-1, 1:-1]
-            elif self.grid.ndim == 3:
-                return padded_ret_array[1:-1, 1:-1, 1:-1]
-            else:  # pragma: no cover
-                raise RuntimeError("Caching error. Please report this.")
+        if boundaries is None:
+            boundaries = {}
+        if weight_field_boundaries is None:
+            if boundaries and weight_field is not None:
+                raise TypeError(
+                    "weight_field_boundaries keyword argument is "
+                    "required with weight_field and boundaries"
+                )
+            weight_field_boundaries = {}
+        elif weight_field is None:
+            warnings.warn(
+                "weight_field_boundaries will not be used "
+                "as no weight_field was specified",
+                stacklevel=2,
+            )
+
+        self._sanitize_boundaries(boundaries)
+        self._sanitize_boundaries(weight_field_boundaries)
 
         known_methods: dict[DepositionMethod, list[DepositionMethodT]] = {
-            DepositionMethod.PARTICLE_IN_CELL: [
-                _deposit_pic_1D,
-                _deposit_pic_2D,
-                _deposit_pic_3D,
+            DepositionMethod.NEAREST_GRID_POINT: [
+                _deposit_ngp_1D,
+                _deposit_ngp_2D,
+                _deposit_ngp_3D,
             ],
             DepositionMethod.CLOUD_IN_CELL: [
                 _deposit_cic_1D,
@@ -475,16 +538,34 @@ class Dataset(ValidatorMixin):
         if mkey not in known_methods:
             raise NotImplementedError(f"method {method} is not implemented yet")
 
-        field = np.array(self.particles.fields[particle_field_key])
+        field = self.particles.fields[particle_field_key]
         padded_ret_array = np.zeros(self.grid._padded_shape, dtype=field.dtype)
+        if weight_field is not None:
+            wfield = np.array(self.particles.fields[weight_field])
+            wfield_dep = np.zeros(self.grid._padded_shape, dtype=field.dtype)
+        else:
+            wfield = np.ones(0, dtype=field.dtype)
+            wfield_dep = np.ones(1, dtype=field.dtype)
+
         self._setup_host_cell_index(verbose=verbose)
 
         func = known_methods[mkey][self.grid.ndim - 1]
         tstart = monotonic_ns()
+        if weight_field is not None:
+            func(
+                *self._get_padded_cell_edges(),
+                *self._get_3D_particle_coordinates(),
+                wfield,
+                np.ones(0, dtype=field.dtype),
+                self._hci,
+                wfield_dep,
+            )
+
         func(
             *self._get_padded_cell_edges(),
             *self._get_3D_particle_coordinates(),
             field,
+            wfield,
             self._hci,
             padded_ret_array,
         )
@@ -494,13 +575,128 @@ class Dataset(ValidatorMixin):
                 f"Deposited {self.particles.count:.4g} particles in {(tstop-tstart)/1e9:.2f} s"
             )
 
-        # boundary conditions treatment should be performed here
-        # ...
+        if weight_field is not None:
+            self._apply_boundary_conditions(
+                array=wfield_dep,
+                boundaries=weight_field_boundaries,
+                weight_array=None,
+            )
 
-        self._cache[particle_field_key, mkey] = padded_ret_array
-        return self.deposit(
-            particle_field_key,
-            method=method,
-            verbose=verbose,
-            return_ghost_padded_array=return_ghost_padded_array,
-        )
+            self._apply_boundary_conditions(
+                array=padded_ret_array,
+                boundaries=boundaries,
+                weight_array=wfield_dep,
+            )
+        else:
+            self._apply_boundary_conditions(
+                array=padded_ret_array,
+                boundaries=boundaries,
+                weight_array=None,
+            )
+
+        padded_ret_array /= wfield_dep
+
+        if return_ghost_padded_array:
+            return padded_ret_array
+        elif self.grid.ndim == 1:
+            return padded_ret_array[1:-1]
+        elif self.grid.ndim == 2:
+            return padded_ret_array[1:-1, 1:-1]
+        elif self.grid.ndim == 3:
+            return padded_ret_array[1:-1, 1:-1, 1:-1]
+        else:  # pragma: no cover
+            raise RuntimeError("Caching error. Please report this.")
+
+    def _sanitize_boundaries(self, boundaries: dict[Name, tuple[Name, Name]]) -> None:
+        if self.grid is None:  # pragma: no cover
+            raise RuntimeWarning("Something took a wrong turn, please report this")
+
+        for ax in self.grid.axes:
+            boundaries.setdefault(ax, ("open", "open"))
+        for axk in boundaries:
+            if axk not in self.grid.axes:
+                raise ValueError(
+                    f"Got invalid ax key {axk!r}, expected any of {self.grid.axes!r}"
+                )
+        for bound in boundaries.values():
+            if (
+                (not isinstance(bound, tuple))
+                or len(bound) != 2
+                or not all(isinstance(_, str) for _ in bound)
+            ):
+                raise TypeError(f"Expected a 2-tuple of strings, got {bound!r}")
+
+            for b in bound:
+                if b not in self.boundary_recipes:
+                    raise ValueError(f"Unknown boundary type {b!r}")
+
+    def _apply_boundary_conditions(
+        self,
+        array: RealArray,
+        boundaries: dict[Name, tuple[Name, Name]],
+        weight_array: RealArray | None,
+    ) -> None:
+        if self.grid is None:  # pragma: no cover
+            raise RuntimeWarning("Something took a wrong turn, please report this")
+
+        axes = list(self.grid.axes)
+        for ax, bv in boundaries.items():
+            # in some applications, it is *crucial* that boundaries be applied
+            # in a specific order, so we take advantage of the fact that dictionaries
+            # are ordered in modern Python and loop over user-specified constraints
+            # as a way to *expose* ordering.
+            iax = axes.index(ax)
+            bcs = tuple(self.boundary_recipes[key] for key in bv)
+            for side, bc in zip(("left", "right"), bcs):
+                side = cast(Literal["left", "right"], side)
+                active_index: int = 1 if side == "left" else -2
+                same_side_active_layer_idx = [slice(None)] * self.grid.ndim
+                same_side_active_layer_idx[
+                    iax
+                ] = active_index  # type:ignore [call-overload]
+
+                same_side_ghost_layer_idx = [slice(None)] * self.grid.ndim
+                # f(-2)=-1, f(1)=0
+                same_side_ghost_layer_idx[iax] = -(  # type:ignore [call-overload]
+                    (active_index + 1) % 2
+                )
+
+                opposite_side_active_layer_idx = [slice(None)] * self.grid.ndim
+                # f(-2)=1, f(1)=-2
+                opposite_side_active_layer_idx[iax] = (  # type:ignore [call-overload]
+                    1 if active_index == -2 else -2
+                )
+
+                opposite_side_ghost_layer_idx = [slice(None)] * self.grid.ndim
+                # f(-2)=0, f(1)=-1
+                opposite_side_ghost_layer_idx[iax] = -(  # type:ignore [call-overload]
+                    active_index % 2
+                )
+
+                if weight_array is None:
+                    ONE = np.ones(1, dtype=array.dtype)
+                    array[tuple(same_side_active_layer_idx)] = bc(
+                        array[tuple(same_side_active_layer_idx)],
+                        array[tuple(same_side_ghost_layer_idx)],
+                        array[tuple(opposite_side_active_layer_idx)],
+                        array[tuple(opposite_side_ghost_layer_idx)],
+                        ONE,
+                        ONE,
+                        ONE,
+                        ONE,
+                        side,
+                        self.metadata,
+                    )
+                else:
+                    array[tuple(same_side_active_layer_idx)] = bc(
+                        array[tuple(same_side_active_layer_idx)],
+                        array[tuple(same_side_ghost_layer_idx)],
+                        array[tuple(opposite_side_active_layer_idx)],
+                        array[tuple(opposite_side_ghost_layer_idx)],
+                        weight_array[tuple(same_side_active_layer_idx)],
+                        weight_array[tuple(same_side_ghost_layer_idx)],
+                        weight_array[tuple(opposite_side_active_layer_idx)],
+                        weight_array[tuple(opposite_side_ghost_layer_idx)],
+                        side,
+                        self.metadata,
+                    )
